from dark.database import getDatabaseConnection


class LineageFetcher(object):
    """
    Provide access to the NCBI taxonomy database so we can retrieve the lineage
    of title sequences hit by BLAST.
    """
    def __init__(self, db=None, cursor=None):
        self._db = db or getDatabaseConnection()
        self._cursor = cursor or self._db.cursor()
        self._cache = {}

    def lineage(self, title):
        """
        Get lineage information from the taxonomy database for a given title.

        @param title: A C{str} sequence title (e.g., from a BLAST hit). Of the
            form 'gi|63148399|gb|DQ011818.1| Description...'. It is the gi
            number (63148399 in this example) that is looked up in the taxonomy
            database.
        @return: A C{list} of the taxonomic categories of the title. Each list
            element is an (C{int}, C{str}) 2-tuple, giving a taxonomy id and
            a scientific name. The first element in the list will correspond to
            C{title}, and each successive element is the parent of the
            preceeding one. If no taxonomy is found, the returned list will be
            empty.
        """
        if title in self._cache:
            return self._cache[title]

        lineage = []
        gi = int(title.split('|')[1])
        query = 'SELECT taxID from gi_taxid where gi = %d' % gi

        try:
            while True:
                self._cursor.execute(query)
                taxID = self._cursor.fetchone()[0]
                if taxID == 1:
                    break
                query = 'SELECT name from names where taxId = %s' % taxID
                self._cursor.execute(query)
                scientificName = self._cursor.fetchone()[0]
                lineage.append((taxID, scientificName))
                # Move up to the parent.
                query = ('SELECT parent_taxID from nodes where taxID = %s' %
                         taxID)
        except TypeError:
            lineage = []

        self._cache[title] = lineage
        return lineage

    def close(self):
        """
        Close the database connection and render self invalid. Any subsequent
        re-use of self will raise an error.
        """
        self._cursor.close()
        self._db.close()
        self._cursor = self._db = self._cache = None



#!/usr/bin/env python

"""
Given a set of FASTA sequence identifiers from sys.argv and/or in a file, read
FASTA from stdin, and print FASTA to stdout for the given sequence ids.
"""

from __future__ import print_function

import sys
import argparse

from Bio import SeqIO


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Extract a subset of FASTA reads by id',
        epilog='Given a set of FASTA sequence identifiers from sys.argv '
        'or in a file, read FASTA from stdin, and print FASTA to stdout '
        'for the given sequence ids.')

    parser.add_argument(
        'ids', default=None, nargs='*', help='Wanted read ids.')

    parser.add_argument(
        '--readIdFile', default=None, help='A file of wanted read ids.')

    args = parser.parse_args()
    wanted = set()

    if args.ids:
        wanted.update(args.ids)

    if args.readIdFile is not None:
        with open(args.readIdFile) as fp:
            for line in fp:
                wanted.add(line[:-1])

    found = []

    if wanted:
        for seq in SeqIO.parse(sys.stdin, 'fasta'):
            if seq.description in wanted:
                wanted.remove(seq.description)
                found.append(seq)

        if found:
            SeqIO.write(found, sys.stdout, 'fasta')

        print('Found %d sequences.' % len(found), file=sys.stderr)

        if wanted:
            print('WARNING: %d sequence%s not found: %s' % (
                len(wanted), '' if len(wanted) == 1 else 's were',
                ' '.join(wanted)), file=sys.stderr)
    else:
        # No wanted ids were given.
        parser.print_help(sys.stderr)
        sys.exit(1)



#!/usr/bin/env python

from __future__ import print_function

from dark.summarize import summarizeReads
import sys


if len(sys.argv) != 2:
    print('Usage: %s file.fasta / file.fastq' % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    filename = sys.argv[1]
    if filename.lower().endswith('q'):
        filetype = 'fastq'
    else:
        # Assume anything not ending with a 'q' is fasta.
        filetype = 'fasta'

    result = summarizeReads(filename, filetype)

    print('Number of reads:', result['read_number'])
    print('Total length: %s bases' % result['total_length'])
    print('The average read length: %s bases' % result['average_length'])
    print('Longest read:', result['max_length'])
    print('Shortest read:', result['min_length'])
    print('Median length:', result['median_length'])

    for base in sorted(result['base_counts']):
        count = result['base_counts'][base]
        print('%s: Total: %s; Average per read: %s' % (
            base, count, count / result['read_number']))



#!/usr/bin/env python

"""
Aligns each sequence in fileHandle1 to each sequence in fileHandle2.
"""

from __future__ import print_function

from dark.local_align import LocalAlignment
import sys
from Bio import SeqIO

if len(sys.argv) != 3 or len(sys.argv) != 8:
    print('Usage: %s filename1, filename2, (matchScore, '
          'mismatchScore, gapOpenScore, gapExtendScore, '
          'gapExtendDecay)'
          % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    fileHandle1 = sys.argv[1]
    fileHandle2 = sys.argv[2]
    if len(sys.argv) == 3:
        matchScore = 1
        mismatchScore = -1
        gapOpenScore = -1
        gapExtendScore = -1
        gapExtendDecay = 0.0
    else:
        matchScore = int(sys.argv[3])
        mismatchScore = int(sys.argv[4])
        gapOpenScore = int(sys.argv[5])
        gapExtendScore = int(sys.argv[6])
        gapExtendDecay = float(sys.argv[7])
    for record in SeqIO.parse(fileHandle1, 'fasta'):
        input1 = record
        for record in SeqIO.parse(fileHandle2, 'fasta'):
            input2 = record
            alignment = LocalAlignment(input1, input2, match=matchScore,
                                       mismatch=mismatchScore,
                                       gap=gapOpenScore,
                                       gapExtend=gapExtendScore,
                                       gapDecay=gapExtendDecay)
            print(alignment.createAlignment())



#!/usr/bin/env python

"""
Given a JSON BLAST output file, a FASTA sequence file, and interesting
criteria, write a FASTA file (to sys.stdout) of the reads that match the
criteria.

Run with --help for help.
"""

from __future__ import print_function

import sys
import argparse

from dark.reads import Reads
from dark.fasta import FastaReads
from dark.titles import TitlesAlignments
from dark.blast.alignments import BlastReadsAlignments


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Extract FASTA from matching BLAST hits.',
        epilog='Given JSON BLAST output files, a FASTA sequence file, '
        'and filtering criteria, produce a FASTA file on stdout '
        'containing the reads that match the criteria.'
    )

    # Args for the JSON BLAST and FASTA files.
    parser.add_argument(
        'json', metavar='BLAST-JSON-file', nargs='+',
        help='the JSON file of BLAST output.')

    parser.add_argument(
        '--fasta', metavar='FASTA-file', required=True,
        help='the FASTA file of sequences that were given to BLAST.')

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        '--minStart', type=int, default=None,
        help='Reads that start before this subject offset should not be '
        'shown.')

    parser.add_argument(
        '--maxStop', type=int, default=None,
        help='Reads that end after this subject offset should not be shown.')

    parser.add_argument(
        '--oneAlignmentPerRead', default=False, action='store_true',
        help='If C{True}, only keep the best alignment for each read.')

    parser.add_argument(
        '--scoreCutoff', type=float, default=None,
        help=('A float score. Matches with scores worse than '
              'this will be ignored.'))

    parser.add_argument(
        '--maxHspsPerHit', type=int, default=None,
        help='A numeric max number of HSPs to show for each hit on hitId.')

    parser.add_argument(
        '--whitelist', nargs='+', default=None,
        help='sequence titles that should be whitelisted')

    parser.add_argument(
        '--blacklist', nargs='+', default=None,
        help='sequence titles that should be blacklisted')

    parser.add_argument(
        '--titleRegex', default=None,
        help='a regex that sequence titles must match.')

    parser.add_argument(
        '--negativeTitleRegex', default=None,
        help='a regex that sequence titles must not match.')

    parser.add_argument(
        '--truncateTitlesAfter', default=None,
        help='a string that titles will be truncated beyond. If a truncated '
        'title has already been seen, that title will be skipped.')

    parser.add_argument(
        '--minSequenceLen', type=int, default=None,
        help='sequences of lesser length will be elided.')

    parser.add_argument(
        '--maxSequenceLen', type=int, default=None,
        help='sequences of greater length will be elided.')

    parser.add_argument(
        '--taxonomy', default=None,
        help='a string of the taxonomic group on which should be '
        'filtered. eg "Vira" will filter on viruses.')

    # Args for filtering on TitlesAlignments.
    parser.add_argument(
        '--minMatchingReads', type=int, default=None,
        help='sequences that are matched by fewer reads will be elided.')

    parser.add_argument(
        '--minMedianScore', type=float, default=None,
        help='sequences that are matched with a median score that is '
        'worse will be elided.')

    parser.add_argument(
        '--withScoreBetterThan', type=float, default=None,
        help='sequences that are matched without at least one score '
        'at least this good will be elided.')

    parser.add_argument(
        '--minNewReads', type=float, default=None,
        help='The fraction of its reads by which a new read set must differ '
        'from all previously seen read sets in order to be considered '
        'acceptably different.')

    args = parser.parse_args()
    reads = FastaReads(args.fasta)
    readsAlignments = BlastReadsAlignments(reads, args.json)

    # Convert white/blacklists lists to sets.
    if args.whitelist is not None:
        args.whitelist = set(args.whitelist)
    if args.blacklist is not None:
        args.blacklist = set(args.blacklist)

    readsAlignments.filter(
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart,
        maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        whitelist=args.whitelist,
        blacklist=args.blacklist,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter,
        taxonomy=args.taxonomy)

    reads = Reads()

    if (args.minMatchingReads is None and args.minMedianScore is None and
            args.withScoreBetterThan is None and args.minNewReads is None):
        # No need to collect into titles, just get the read ids from
        # the matching alignments.
        for readAlignment in readsAlignments:
            reads.add(readAlignment.read)
    else:
        # We need to collect alignments into titles.
        titlesAlignments = TitlesAlignments(readsAlignments).filter(
            minMatchingReads=args.minMatchingReads,
            minMedianScore=args.minMedianScore,
            withScoreBetterThan=args.withScoreBetterThan,
            minNewReads=args.minNewReads)

        for titleAlignments in titlesAlignments.values():
            for alignment in titleAlignments.alignments:
                reads.add(alignment.read)

    reads.save(sys.stdout)
    print('Found %d matching reads.' % len(reads), file=sys.stderr)



#!/usr/bin/env python

"""
Convert an e-value to a bit score.

Example usage:

$ e-value-to-bit-score.py --dbSize 168142520 --dbSequenceCount 5660 \
  --queryLength 111 --lengthAdjustment 26 --eValue 0.0813077725194
37.3537
"""

from __future__ import print_function

import argparse

from dark.blast.score import eValueToBitScore


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Convert a bit score to an e-value.')

    parser.add_argument(
        '--eValue', type=float, required=True, help='The e-value to convert.')

    parser.add_argument(
        '--dbSize', type=int, required=True,
        help='The total number of bases in the sequence database.')

    parser.add_argument(
        '--dbSequenceCount', type=int, required=True,
        help='The number of sequences in the database.')

    parser.add_argument(
        '--queryLength', type=int, required=True,
        help='The length of the query sequence.')

    parser.add_argument(
        '--lengthAdjustment', type=int, required=True,
        help='The length adjustment.')

    args = parser.parse_args()
    print(eValueToBitScore(args.eValue, args.dbSize, args.dbSequenceCount,
                           args.queryLength, args.lengthAdjustment))
